@@ -4,17 +4,19 @@
 (*                                                                     *)
 (*                Daniel de Rauglaudre, INRIA Rocquencourt             *)
 (*                                                                     *)
-(*  Copyright 2001-2008 Institut National de Recherche en Informatique *)
+(*  Copyright 2001-2010 Institut National de Recherche en Informatique *)
 (*  et Automatique.  Distributed only by permission.                   *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: ledit.ml,v 1.57 2008/01/23 09:33:59 deraugla Exp $ *)
+(* $Id: ledit.ml,v 1.67 2010-11-17 16:30:50 deraugla Exp $ *)
 
 #load "pa_local.cmo";
 #load "pa_def.cmo";
 #load "pa_fstream.cmo";
 
+open Printf;
+
 type encoding = [ Ascii | Iso_8859 | Utf_8 ];
 
 module A :
@@ -195,7 +197,8 @@ module A :
         value input_line ic =
           let s = input_line ic in
           match encoding.val with
-          [ Ascii | Iso_8859 -> of_ascii s
+          [ Ascii | Iso_8859 ->
+              Array.init (String.length s) (fun i -> String.make 1 s.[i])
           | Utf_8 ->
               loop [] 0 where rec loop list i =
                 if i >= String.length s then Array.of_list (List.rev list)
@@ -232,6 +235,7 @@ type command =
   | Beginning_of_history
   | Beginning_of_line
   | Capitalize_word
+  | Complete_file_name
   | Delete_char
   | Delete_char_or_end_of_file
   | Downcase_word
@@ -401,6 +405,7 @@ value init_default_commands kb =
      ("\\C-d", Delete_char_or_end_of_file);
      ("\\C-h", Backward_delete_char);
      ("\\177", Backward_delete_char);
+     ("\\C-i", Complete_file_name);
      ("\\C-t", Transpose_chars);
      ("\\C-q", Quoted_insert);
      ("\\C-k", Kill_line);
@@ -432,7 +437,7 @@ value init_default_commands kb =
      ("\\e[H", Beginning_of_line);	(* Home *)
      ("\\e[F", End_of_line);		(* End *)
      ("\\e[5~", Previous_history);	(* Page Up *)
-     ("\\e[6~", Next_history);		(* Page Down *)
+     ("\\e[6~", Next_history);          (* Page Down *)
      ("\\e[2H", Beginning_of_history);	(* Shift Home *)
      ("\\e[2F", End_of_history);	(* Shift End *)
      ("\\e[OA", Previous_history);
@@ -440,17 +445,17 @@ value init_default_commands kb =
      ("\\e[OD", Backward_char);
      ("\\e[OH", Beginning_of_line) ::
      if meta_as_escape.val then
-       [("\\M-f", Forward_word);
-        ("\\M-b", Backward_word);
+       [("\\M-b", Backward_word);
+        ("\\M-c", Capitalize_word);
+        ("\\M-d", Kill_word);
+        ("\\M-f", Forward_word);
+        ("\\M-l", Downcase_word);
+        ("\\M-u", Upcase_word);
         ("\\M-<", Beginning_of_history);
         ("\\M->", End_of_history);
-        ("\\M-c", Capitalize_word);
-        ("\\M-u", Upcase_word);
-        ("\\M-l", Downcase_word);
-        ("\\M-d", Kill_word);
+        ("\\M-/", Expand_abbrev);
         ("\\M-\\C-h", Backward_kill_word);
-        ("\\M-\\127", Backward_kill_word);
-        ("\\M-/", Expand_abbrev)]
+        ("\\M-\\127", Backward_kill_word)]
      else []]
 ;
 
@@ -524,6 +529,7 @@ value command_of_name = do {
   add "end-of-history" End_of_history;
   add "end-of-line" End_of_line;
   add "expand-abbrev" Expand_abbrev;
+  add "complete-file-name" Complete_file_name;
   add "forward-char" Forward_char;
   add "forward-word" Forward_word;
   add "interrupt" Interrupt;
@@ -536,6 +542,7 @@ value command_of_name = do {
   add "quoted-insert" Quoted_insert;
   add "redraw-current-line" Redraw_current_line;
   add "reverse-search-history" Reverse_search_history;
+  add "self-insert" Self_insert;
   add "suspend" Suspend;
   add "transpose-chars" Transpose_chars;
   add "unix-line-discard" Unix_line_discard;
@@ -599,7 +606,9 @@ type state =
     last_comm : mutable command;
     histfile : mutable option out_channel;
     history : mutable Cursor.t A.String.t;
-    abbrev : mutable option abbrev_data }
+    abbrev : mutable option abbrev_data;
+    complete_fn : mutable int;
+    complete_fn_screen : mutable int}
 ;
 
 value eval_comm s st =
@@ -666,7 +675,7 @@ value bell () = do { prerr_string "\007"; flush stderr };
 value saved_tcio =
   try Unix.tcgetattr Unix.stdin with
   [ Unix.Unix_error _ _ _ -> do {
-      Printf.eprintf "Error: standard input is not a terminal\n";
+      eprintf "Error: standard input is not a terminal\n";
       flush stderr;
       exit 1
     } ]
@@ -1085,6 +1094,9 @@ value reverse_search_history st =
             bell ();
             initial_str
           }
+        | Self_insert ->
+            let str = A.String.concat str (A.String.of_char c) in
+            incr_search (find_line (cnt, fstr) str) str
         | _ -> fstr ]
     | Some None ->
         if s = "\027" then fstr
@@ -1168,9 +1180,175 @@ value expand_abbrev st abbrev = do {
   update_output st
 };
 
+value start_with s s_ini =
+  let len = String.length s_ini in
+  String.length s >= len && String.sub s 0 len = s_ini
+;
+
+value insert_string st s = do {
+  String.iter
+     (fun c -> do {
+        insert_char st (A.Char.of_ascii c);
+        st.line.cur := st.line.cur + 1
+      }) s
+};
+
+value is_directory fn =
+  try Sys.is_directory fn with [ Sys_error _ -> False ]
+;
+
+value print_file_list st max_flen nb_by_line dirname files = do {
+  loop 0 files where rec loop n =
+    fun
+    [ [file :: files] -> do {
+        if n = nb_by_line then do {
+          prerr_endline "";
+          loop 0 [file :: files]
+        }
+        else do {
+          prerr_string file;
+          let fn = Filename.concat dirname file in
+          if is_directory fn then prerr_string Filename.dir_sep
+          else prerr_string " ";
+          if n < nb_by_line - 1 then do {
+            prerr_string
+              (String.make (max_flen + 1 - String.length file) ' ');
+          }
+          else ();
+          loop (n + 1) files;
+        }
+      }
+    | [] -> () ];
+};
+
+value rev_take n list =
+  loop [] n list where rec loop rev_list n =
+    fun
+    [ [x :: l] ->
+        if n = 0 then (rev_list, [x :: l])
+        else loop [x :: rev_list] (n - 1) l
+    | [] -> (rev_list, []) ]
+;
+
+value max_lines_in_screen = ref 24;
+
+value complete_file_name st = do {
+  let s =
+    loop "" (st.line.cur - 1) where rec loop s i =
+      if i < 0 then s
+      else
+        match A.Char.to_ascii (A.String.get st.line.buf i) with
+        [ Some c ->
+            match c with
+            [ 'a'..'z' | 'A'..'Z' | '0'..'9' | '.' | '_' | '-' | '/' | '#' |
+              '~' ->
+                loop (String.make 1 c ^ s) (i - 1)
+            | _ -> s ]
+        | None -> s ]
+  in
+  let dirname = Filename.dirname s in
+  let basename = Filename.basename s in
+  if is_directory dirname then do {
+    let files = Array.to_list (Sys.readdir dirname) in
+    let (files, basename) =
+      if s = "" then (files, "")
+      else if s = Filename.current_dir_name ^ Filename.dir_sep then
+        (files, "")
+      else if s = Filename.parent_dir_name ^ Filename.dir_sep then
+        (files, "")
+      else if
+        basename = Filename.current_dir_name &&
+        s <> Filename.concat dirname basename
+      then (files, "")
+      else
+        ([Filename.current_dir_name; Filename.parent_dir_name :: files],
+         basename)
+    in
+    let files = List.filter (fun fn -> start_with fn basename) files in
+    let files = List.sort compare files in
+    let max_flen =
+      List.fold_left (fun max_len fn -> max max_len (String.length fn)) 0
+        files
+    in
+    let nb_by_line = max 1 ((max_len.val + 2) / (max_flen + 2)) in
+    let nfiles = List.length files in
+    let nb_lines = (nfiles + nb_by_line - 1) / nb_by_line in
+    let mlis = max_lines_in_screen.val - 1 in
+    let nb_screens = (nb_lines + mlis - 1) / mlis in
+    let (files_to_display, screen_nb) =
+      if nb_screens <= 1 then (files, 0)
+      else do {
+        let nb_files_by_screen = nb_by_line * mlis in
+        let screen_nb = st.complete_fn_screen mod nb_screens in
+        let rev_files =
+          loop screen_nb files where rec loop n files =
+            let (rev_files, rest) = rev_take nb_files_by_screen files in
+            if n = 0 then rev_files
+            else loop (n - 1) rest
+        in
+        st.complete_fn_screen := st.complete_fn_screen + 1;
+        (List.rev
+           (if screen_nb < nb_screens - 1 then  ["..." :: rev_files]
+            else rev_files),
+         screen_nb)
+      }
+    in
+    match files with
+    [ [] -> ()
+    | [fname] -> do {
+        let len = String.length basename in
+        let s = String.sub fname len (String.length fname - len) in
+        insert_string st s;
+        let fn = Filename.concat dirname fname in
+        if is_directory fn then insert_string st Filename.dir_sep
+        else insert_string st " ";
+        update_output st
+      }
+    | [file :: files] -> do {
+        let common =
+          loop file files where rec loop common =
+            fun
+            [ [file :: files] ->
+                if start_with file common then loop common files
+                else if common = "" then ""
+                else
+                  loop (String.sub common 0 (String.length common - 1))
+                    [file :: files]
+            | [] -> common ]
+        in
+        let len = String.length basename in
+        if String.length common > len then do {          
+          let s = String.sub common len (String.length common - len) in
+          insert_string st s;
+          update_output st;
+        }
+        else ();
+        put_newline st;
+        st.od.cur := 0;
+        st.od.len := 0;
+        prerr_string "*** files";
+        if nb_screens > 1 then eprintf " (%d/%d)" (screen_nb + 1) nb_screens
+        else ();
+        prerr_endline " ***";
+        print_file_list st max_flen nb_by_line dirname files_to_display;
+        put_newline ();
+        flush stderr;
+        update_output st;
+      } ]
+  }
+  else ()
+};
+
 value rec update_line st comm c = do {
   let abbrev = st.abbrev in
   st.abbrev := None;
+  if comm = Complete_file_name then do {
+    st.complete_fn := st.complete_fn + 1
+  }
+  else do {
+    st.complete_fn := 0;
+    st.complete_fn_screen := 0;
+  };
   match comm with
   [ Beginning_of_line ->
       if st.line.cur > 0 then do {st.line.cur := 0; update_output st} else ()
@@ -1281,6 +1459,7 @@ value rec update_line st comm c = do {
       update_output st
     }
   | Expand_abbrev -> expand_abbrev st abbrev
+  | Complete_file_name -> complete_file_name st
   | Redraw_current_line -> do {
       put_newline st;
       st.od.cur := 0;
@@ -1377,14 +1556,14 @@ local st =
    last_line = A.String.empty; istate = Normal ""; shift = 0;
    init_kb = None; total_kb = None; cut = A.String.empty;
    last_comm = Accept_line; histfile = None; history = Cursor.create ();
-   abbrev = None}
+   abbrev = None; complete_fn = 0; complete_fn_screen = 0}
 in
 value edit_line () = do {
   let rec edit_loop () = do {
     let c = A.Char.read () in
     if trace_sequences.val then do {
       put_newline st;
-      Printf.eprintf "<%s>\n" (String.escaped (A.Char.to_string c));
+      eprintf "<%s>\n" (String.escaped (A.Char.to_string c));
       st.od.cur := 0;
       st.od.len := 0;
       update_output st
